/*
        factory.h -- polymorphic cluster:
            instance factory & class registry
        (C) Copyright 1995  John Webster Small
        All rights reserved

        DERIVED must have a default constructor or
        define the Die_traits<> template to implement a
        default constructor:

        struct Die_traits<DERIVED,ROOT>
        {
            static ROOT * newInstance()
                { return ??? new DERIVED ???  }
        };

*/

#ifndef factory_h
#define factory_h

/* LINTLIBRARY */

#ifndef cl_h
  #include "cl.h"
#endif

#ifndef single_h
  #include "single.h"
#endif

#ifndef typeinfo_h
  #include <typeinfo.h>
  #ifndef typeinfo_h
    #define typeinfo_h
  #endif
#endif

template <class ROOT>
class Machine
{
  public:
    Machine() {}
    virtual ~Machine() {}
    virtual const Type_info& v_typeid() const
        { return typeid(ROOT); }
    virtual ROOT * newInstance() const
        { return 0; }
    int cmp(const char * typeid_name) const
        { return strcmp(v_typeid().name(),
          typeid_name); }
    int cmp(const Machine<ROOT >& m) const
        { return cmp(m.v_typeid().name()); }
    CL_StrongTypingCmp(Machine<ROOT >);
    static int cmpName2Machine
        (const char * typeid_name,
         const Machine<ROOT > * M)
        { return - M->cmp(typeid_name); }
};

template <class DERIVED, class ROOT>
struct Die_traits
{
    static ROOT * newInstance()
        { return new DERIVED; }
};

template <class DERIVED, class ROOT>
class Die : public Machine<ROOT >
{
  public:
    Die() : Machine<ROOT >() {}
    virtual const Type_info& v_typeid() const
        { return typeid(DERIVED); }
    virtual ROOT * newInstance() const
      { return Die_traits<DERIVED,ROOT >::
        newInstance(); }
};

#define static_registerClass(DERIVED,ROOT) \
    static void registerClass() \
    { \
        if (!registered(typeid(DERIVED).name())) \
            registerDL(new Die<DERIVED,ROOT >()); \
    }

template <class ROOT>
class Factory : CL<Machine<ROOT > >
{
  public:
    Factory() : CL<Machine<ROOT > >(CL_DEL) {}
    Machine<ROOT > * registered
        (const char * typeid_name) const;
    Machine<ROOT > * operator[]
        (const char * typeid_name) const
        { return registered(typeid_name); }
    int registerDL(Machine<ROOT > * M)
        { return insUniqueDL(M) != 0; }
    void forget(const char * typeid_name);
    void forgetAll() { allClr(); }
    ROOT * newInstance(const char * typeid_name) const;
};

template <class ROOT>
Machine<ROOT > *
Factory<ROOT >::
    registered(const char * typeid_name) const
{
    if (typeid_name)  {
        unsigned idx;
        return findFirst(typeid_name,idx,
            CLKCCSM(Machine<ROOT >,cmpName2Machine));
    }
    return 0;
}

template <class ROOT>
void Factory<ROOT >::forget(const char * typeid_name)
{
    if (typeid_name)
        if (findFirst(typeid_name,
            CLKCCSM(Machine<ROOT >,cmpName2Machine)))
            (void) clrBackup();
}

template <class ROOT>
ROOT *
Factory<ROOT >::newInstance
    (const char * typeid_name) const
{
    Machine<ROOT > * M = registered(typeid_name);
    return (M? M->newInstance() : 0);
}

template <class ROOT>
class Registry
{
  public:
    Registry() {}
    ~Registry() {}
    static Machine<ROOT > * registered
        (const char * typeid_name);
    static int registerDL(Machine<ROOT > * M);
    static void forget(const char * typeid_name);
    static void forgetAll();
    static ROOT * newInstance
        (const char * typeid_name);
};

template <class ROOT>
Machine<ROOT > *
Registry<ROOT >:: registered (const char * typeid_name)
{
    static Singleton<Registry<ROOT >,
        Factory<ROOT > > registry;
    if (registry.ok())
        return ((Factory<ROOT >&)registry)
            .registered(typeid_name);
    return 0;
}

template <class ROOT>
int Registry<ROOT >:: registerDL(Machine<ROOT > * M)
{
    static Singleton<Registry<ROOT >,
        Factory<ROOT > > registry;
    if (registry)
        return ((Factory<ROOT >&)registry)
            .registerDL(M);
    delete M;
    return 0;
}

template <class ROOT>
void
Registry<ROOT >:: forget(const char * typeid_name)
{
    static Singleton<Registry<ROOT >,
        Factory<ROOT > > registry;
    if (registry.ok())
        ((Factory<ROOT >&)registry)
            .forget(typeid_name);
}

template <class ROOT>
void Registry<ROOT >:: forgetAll()
{
    static Singleton<Registry<ROOT >,
        Factory<ROOT > > registry;
    if (registry.ok())
        ((Factory<ROOT >&)registry).forgetAll();
}

template <class ROOT>
ROOT *
Registry<ROOT >:: newInstance(const char * typeid_name)
{
    static Singleton<Registry<ROOT >,
        Factory<ROOT > > registry;
    if (registry.ok())
        return ((Factory<ROOT >&)registry)
            .newInstance(typeid_name);
    return 0;
}

#define static_RegistryPromote(DERIVED_ROOT, ROOT) \
    static Machine<ROOT > * registered \
        (const char * typeid_name) \
        { return Registry<ROOT >:: \
          registered(typeid_name); } \
    static int registerDL(Machine<ROOT > * M) \
        { return Registry<ROOT >::registerDL(M); } \
    static void forget(const char * typeid_name) \
        { Registry<ROOT >::forget(typeid_name); } \
    static void forgetAll() \
        { Registry<ROOT >::forgetAll(); } \
    static DERIVED_ROOT * newInstance \
        (const char * typeid_name) \
        { return (DERIVED_ROOT *) Registry<ROOT >:: \
          newInstance(typeid_name); }

#endif
